#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import pytest

import disent.registry as R

# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


COUNTS = {
    "DATASETS": 15,
    "SAMPLERS": 8,
    "FRAMEWORKS": 15,
    "RECON_LOSSES": 9,
    "LATENT_HANDLERS": 2,
    "OPTIMIZERS": 30,
    "METRICS": 9,
    "SCHEDULES": 5,
    "MODELS": 8,
    "KERNELS": 2,
}

COUNTS = {
    "DATASETS": 17,
    "SAMPLERS": 8,
    "FRAMEWORKS": 17,
    "RECON_LOSSES": 9,
    "LATENT_HANDLERS": 2,
    "OPTIMIZERS": 30,
    "METRICS": 9,
    "SCHEDULES": 5,
    "MODELS": 8,
    "KERNELS": 2,
}


@pytest.mark.parametrize("registry_key", COUNTS.keys())
def test_registry_loading(registry_key):
    from docs.examples.extend_experiment.code import register_to_disent

    register_to_disent()
    register_to_disent()  # must be able to call more than once!
    # load everything and check the counts
    count = 0
    for example in R.REGISTRIES[registry_key]:
        loaded = R.REGISTRIES[registry_key][example]
        count += 1
    assert count == COUNTS[registry_key], f"invalid count for: {registry_key}"


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
import torch

# ========================================================================= #
# DEBUG                                                                     #
# ========================================================================= #


def replace_arrays_with_shapes(obj):
    """
    recursively replace all arrays of an object
    with their shapes to make debugging easier!
    """
    if isinstance(obj, (torch.Tensor, np.ndarray)):
        return obj.shape
    elif isinstance(obj, dict):
        return {replace_arrays_with_shapes(k): replace_arrays_with_shapes(v) for k, v in obj.items()}
    elif isinstance(obj, list):
        return list(replace_arrays_with_shapes(v) for v in obj)
    elif isinstance(obj, tuple):
        return tuple(replace_arrays_with_shapes(v) for v in obj)
    elif isinstance(obj, set):
        return {replace_arrays_with_shapes(k) for k in obj}
    else:
        return obj


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
from tempfile import TemporaryDirectory
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Tuple

import numpy as np
from PIL import Image

from disent.dataset.data import DiskGroundTruthData
from disent.dataset.data import GroundTruthData
from disent.dataset.util.datafile import DataFileHashed

log = logging.getLogger(__name__)

# ========================================================================= #
# helper                                                                    #
# ========================================================================= #


SPRITES_REPO = "https://github.com/YingzhenLi/Sprites"
SPRITES_REPO_COMMIT_SHA = "3ce4048c5227802bd8f1888e293fd3afdba91c0c"


def fetch_sprite_components() -> Tuple[np.array, np.array]:
    try:
        import git
    except ImportError:
        log.error("GitPython not found! Please install it: `pip install GitPython`")
        exit(1)
    # store files in a temporary directory
    with TemporaryDirectory(suffix="sprites") as temp_dir:
        # clone the files into the temp dir
        log.info(f"Generating sprites data, temporarily cloning: {SPRITES_REPO} to {temp_dir}`")
        repo = git.Repo.clone_from(SPRITES_REPO, temp_dir, no_checkout=True)
        repo.git.checkout(SPRITES_REPO_COMMIT_SHA)
        # get all the components!
        component_sheets: List[np.ndarray] = []
        component_names = ["bottomwear", "topwear", "hair", "eyes", "shoes", "body"]
        for sprites_folder in component_names:
            # append all sprite sheets for the current component
            sheets = []
            for sheet_name in sorted(os.listdir(os.path.join(temp_dir, sprites_folder))):
                img_path = os.path.join(temp_dir, sprites_folder, sheet_name)
                img = np.array(
                    Image.open(img_path).convert("RGBA")
                )  # imageio sometimes doesnt load these all in the same way, we explicitly set the format as RGBA!
                sheets.append(img)
            # append all the sheets
            sheets = np.array(sheets)
            component_sheets.append(sheets)
            # print information
            log.debug(f"{sprites_folder} {sheets.shape} {sheets.dtype}")
    # done!
    return np.array(component_sheets, dtype=object), np.array(component_names, dtype=object)


def save_sprite_components(out_file: str, sheets, names) -> NoReturn:
    # get the path and make the parant dirs
    out_file = os.path.abspath(out_file)
    log.debug(f"saving: {out_file}")
    os.makedirs(os.path.dirname(out_file), exist_ok=True)
    # save the data
    np.savez_compressed(out_file, sheets=sheets, names=names)


def load_sprite_components(in_file: str) -> Tuple[Tuple[np.ndarray, ...], Tuple[str, ...]]:
    dat = np.load(in_file, allow_pickle=True)
    return dat["sheets"], dat["names"]


# ========================================================================= #
# datafiles                                                                 #
# ========================================================================= #


class DataFileSprites(DataFileHashed):
    """
    Download the Sprites GitHub repo and convert it
    to lists of arrays of images that are pickled.
    """

    def _prepare(self, out_dir: str, out_file: str) -> NoReturn:
        sheets, names = fetch_sprite_components()
        save_sprite_components(out_file, sheets=sheets, names=names)


# ========================================================================= #
# dataset_sprites                                                           #
# ========================================================================= #


class SpritesAllData(DiskGroundTruthData):
    """
    Custom version of sprites, with the data obtained from:
    https://github.com/YingzhenLi/Sprites
    """

    name = "sprites"
    factor_names = ("bottomwear", "topwear", "hair", "eyes", "shoes", "body", "action", "rotation", "frame")
    factor_sizes = (7, 7, 10, 5, 3, 7, 5, 4, 6)  # 6_174_000
    img_shape = (64, 64, 3)

    datafile = DataFileSprites(
        file_name="sprites.npz",
        file_hash={"fast": "5c739a7c8e59a20ec34439213036993a", "full": "189baaca306cd305f51291c5decad18d"},
    )
    datafiles = (datafile,)

    # 21x13 tiles in size @ 64x64 px gives 1344x832 px total or 273 tiles total
    SPRITE_SHEET_ACTIONS = {
        "spellcard": {"back": range(0, 7), "left": range(13, 20), "front": range(26, 33), "right": range(39, 46)},
        "thrust": {"back": range(52, 60), "left": range(65, 73), "front": range(78, 86), "right": range(91, 99)},
        "walk": {"back": range(104, 113), "left": range(117, 126), "front": range(130, 139), "right": range(143, 152)},
        "slash": {"back": range(156, 162), "left": range(169, 175), "front": range(182, 188), "right": range(195, 201)},
        "shoot": {"back": range(208, 221), "left": range(221, 234), "front": range(234, 247), "right": range(247, 260)},
        "hurt": {"front": range(260, 266)},
    }

    def __init__(self, data_root: Optional[str] = None, prepare: bool = False, transform=None):
        super().__init__(data_root=data_root, prepare=prepare, transform=transform)
        # load the data
        dat = np.load(os.path.join(self._data_dir, self.datafile.out_name), allow_pickle=True)
        self._names = dat["names"]
        self._sheets = dat["sheets"]

    def sample_random_frames(self, alpha: bool = True, combined: bool = False) -> List[np.ndarray]:
        return self.get_frames(idx=self.sample_indices(), alpha=alpha, combined=combined)

    def get_frames(self, idx: Optional[int], alpha: bool = True, combined: bool = False) -> List[np.ndarray]:
        *sheet_idxs, act, rot, frame = self.idx_to_pos(idx)
        # extract the individual sheets
        sheets = (sheets[i] for i, sheets in zip(sheet_idxs, self._sheets))
        # extract the individual tiles
        y, x = act * 4 + rot, frame
        # return individual frames
        frames = [sheet[y * 64 : (y + 1) * 64, x * 64 : (x + 1) * 64, :] for sheet in sheets]
        # return the combined frame
        if combined:
            frames.append(self.combine_frames(frames, alpha=True))
        # return the final result
        if not alpha:
            return [frame[:, :, :3] for frame in frames]
        return frames

    def combine_frames(self, frames_rgba, alpha: bool = True):
        # combine the frames in order
        img = np.zeros((64, 64, 4) if alpha else (64, 64, 3), dtype="uint8")
        for i in [5, 4, 0, 1, 3, 2]:  # body, shoes, bottomwear, topwear, eyes, hair
            mask = frames_rgba[i][:, :, 3] > 128
            img[:, :, 0][mask] = frames_rgba[i][:, :, 0][mask]
            img[:, :, 1][mask] = frames_rgba[i][:, :, 1][mask]
            img[:, :, 2][mask] = frames_rgba[i][:, :, 2][mask]
            # union for the mask
            if alpha:
                img[:, :, 3][mask] = 255
        # return the result
        return img

    def _get_observation(self, idx):
        frames = self.get_frames(idx, alpha=True, combined=False)
        img = self.combine_frames(frames, alpha=False)
        return img


class SpritesPartialData(GroundTruthData):
    """
    Same as `SpritesOrigData` but the backwards facing `rotation` is removed because it
    hides some of the other ground-truth factors. Similarely, the 5th `bottomwear` and `topwear`
    are removed too because they are missing entries.
    """

    img_shape = SpritesAllData.img_shape
    factor_names = SpritesAllData.factor_names
    factor_sizes = (6, 6, 10, 5, 3, 7, 5, 3, 6)  # 3_402_000

    def __init__(self, data_root: Optional[str] = None, prepare: bool = False, transform=None):
        super().__init__(transform=transform)
        self._sprites = SpritesAllData(data_root=data_root, prepare=prepare, transform=None)

    def _offset_idx(self, idx: int) -> int:
        pos = self.idx_to_pos(idx)
        # convert to orig state space
        if pos[0] >= 5:
            pos[0] += 1  # no missing pants
        if pos[1] >= 5:
            pos[1] += 1  # no missing shirt
        if pos[7] >= 0:
            pos[7] += 1  # no backwards facing man
        idx = self._sprites.pos_to_idx(pos)
        # index in orig state space
        return idx

    def sample_random_frames(self, alpha: bool = True, combined: bool = False) -> List[np.ndarray]:
        return self._sprites.sample_random_frames(alpha=alpha, combined=combined)

    def get_frames(self, idx: int, alpha: bool = True, combined: bool = False) -> List[np.ndarray]:
        return self._sprites.get_frames(idx=self._offset_idx(idx), alpha=alpha, combined=combined)

    def combine_frames(self, frames_rgba, alpha: bool = True):
        return self._sprites.combine_frames(frames_rgba=frames_rgba, alpha=alpha)

    def _get_observation(self, idx):
        return self._sprites._get_observation(self._offset_idx(idx))


class SpritesData(SpritesPartialData):
    """
    Same as `SpritesOrigData` but the backwards facing `rotation` is removed because it
    hides some of the other ground-truth factors.
    """

    factor_sizes = (7, 7, 10, 5, 3, 7, 5, 3, 6)  # 4_630_500

    def _offset_idx(self, idx: int) -> int:
        pos = self.idx_to_pos(idx)
        # convert to orig state space
        if pos[7] >= 0:
            pos[7] += 1  # no backwards facing man
        idx = self._sprites.pos_to_idx(pos)
        # index in orig state space
        return idx


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


if __name__ == "__main__":

    def main():
        import matplotlib.pyplot as plt
        from tqdm import tqdm

        from disent.util.visualize.plot import plt_imshow
        from disent.util.visualize.vis_util import make_image_grid

        data = SpritesAllData(prepare=True)

        # show frames and combined frames
        frames = data.sample_random_frames(True, True)
        plt_imshow(make_image_grid(frames, num_cols=-1), show=True)
        plt_imshow(make_image_grid([f[:, :, 0:3] for f in frames], num_cols=-1), show=True)
        plt_imshow(make_image_grid([f[:, :, 3:4] for f in frames], num_cols=-1), show=True)

        # plot factor traversals
        factors, indices = data.sample_random_factor_traversal_grid(num=10, return_indices=True)
        plt_imshow(make_image_grid([data[i] for i in indices.flatten()], num_cols=10, pad=4))
        plt.show()

        # check speeds
        for i, _ in enumerate(tqdm(SpritesAllData(prepare=True))):  # ~7443 it/s
            if i > 25000:
                break
        for i, _ in enumerate(tqdm(SpritesData(prepare=True))):  # ~6355 it/s
            if i > 25000:
                break

    main()



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
import re
from pathlib import Path
from typing import Optional
from typing import Tuple
from typing import Union

log = logging.getLogger(__name__)


# ========================================================================= #
# PATH HELPERS                                                              #
# ========================================================================= #


_EXPERIMENT_SEP = "_"
_EXPERIMENT_RGX = re.compile(f"^([0-9]+)({_EXPERIMENT_SEP}.+)?$")


def get_max_experiment_number(root_dir: str, return_path: bool = False) -> Union[int, Tuple[int, Optional[str]]]:
    """
    Get the next experiment number in the specified directory. Experiment directories
    all start with a numerical value.
    - eg. "1", "00002", "3_name", "00042_name" are all valid subdirectories.
    - eg. "name", "name_1", "name_00001", "99999_image.png" are all invalid and are
          ignored. Either their name format is wrong or they are a file.

    If all the above directories are all used as an example, then this function will
    return the value 42 corresponding to "00042_name"
    """
    # check the dirs exist
    if not os.path.exists(root_dir):
        raise FileNotFoundError(
            f"The given experiments directory does not exist: {repr(root_dir)} ({repr(os.path.abspath(root_dir))})"
        )
    elif not os.path.isdir(root_dir):
        raise NotADirectoryError(
            f"The given experiments path exists, but is not a directory: {repr(root_dir)} ({repr(os.path.abspath(root_dir))})"
        )
    # linear search over each file in the dir
    max_num, max_path = 0, None
    for file in os.listdir(root_dir):
        # skip if not a directory
        if not os.path.isdir(os.path.join(root_dir, file)):
            continue
        # skip if the file name does not match
        match = _EXPERIMENT_RGX.search(file)
        if not match:
            continue
        # update the maximum number
        num, _ = match.groups()
        num = int(num)
        if num > max_num:
            max_num, max_path = num, file
    # done!
    if return_path:
        return max_num, max_path
    return max_num


_CURRENT_EXPERIMENT_NUM: Optional[int] = None
_CURRENT_EXPERIMENT_DIR: Optional[str] = None


def get_current_experiment_number(root_dir: str) -> int:
    """
    Get the next experiment number from the experiment directory, and cache
    the result for future calls of this function for the current instance of the program.
    - The next time the program is run, this value will differ.

    For example, if the `root_dir` contains the directories: "00001_name", "00041", then
    this function will return the next value which is `42` on all subsequent calls, even
    if a directory for experiment 42 is created during the current program's lifetime.
    """
    global _CURRENT_EXPERIMENT_NUM
    if _CURRENT_EXPERIMENT_NUM is None:
        _CURRENT_EXPERIMENT_NUM = get_max_experiment_number(root_dir, return_path=False) + 1
    return _CURRENT_EXPERIMENT_NUM


def get_current_experiment_dir(root_dir: str, name: Optional[str] = None) -> str:
    """
    Like `get_current_experiment_number` which computes the next experiment number, this
    function computes the next experiment path, which appends a name to the computed number.

    The result is cached for the lifetime of the program, however, on subsequent calls of
    this function, the computed name must always match the original value otherwise an
    error is thrown! This is to prevent experiments with duplicate numbers from being created!
    """
    if name is not None:
        assert Path(name).name == name, f"The given name is not valid: {repr(name)}"
    # make the dirname & normalise the path
    num = get_current_experiment_number(root_dir)
    dir_name = f"{num:05d}{_EXPERIMENT_SEP}{name}" if name else f"{num:05d}"
    exp_dir = os.path.abspath(os.path.join(root_dir, dir_name))
    # cache the experiment name or check against the existing cache
    global _CURRENT_EXPERIMENT_DIR
    if _CURRENT_EXPERIMENT_DIR is None:
        _CURRENT_EXPERIMENT_DIR = exp_dir
    if exp_dir != _CURRENT_EXPERIMENT_DIR:
        raise RuntimeError(
            f"Current experiment directory has already been set: {repr(_CURRENT_EXPERIMENT_DIR)} This does not match what was computed: {repr(exp_dir)}"
        )
    # done!
    return _CURRENT_EXPERIMENT_DIR


def make_current_experiment_dir(root_dir: str, name: Optional[str] = None) -> str:
    """
    Like `get_current_experiment_dir`, but create any of the directories if needed.
    - Both the `root_dir` and the computed subdir for the current experiment will be created.
    """
    root_dir = os.path.abspath(root_dir)
    # make the root directory if it does not exist!
    if not os.path.exists(root_dir):
        log.info(f"root experiments directory does not exist, creating... {repr(root_dir)}")
        os.makedirs(root_dir, exist_ok=True)
    # get the current dir
    current_dir = get_current_experiment_dir(root_dir, name)
    # make the current dir
    if not os.path.exists(current_dir):
        log.info(f"current experiment directory does not exist, creating... {repr(current_dir)}")
        os.makedirs(current_dir, exist_ok=True)
    # done!
    return current_dir


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging

import lightning as L

from disent.util.lightning.callbacks._callbacks_base import BaseCallbackTimed

log = logging.getLogger(__name__)


# ========================================================================= #
# General Callbacks that work for any L.LightningModule                     #
# ========================================================================= #


class LoggerProgressCallback(BaseCallbackTimed):
    def __init__(self, interval: float = 10, log_level: int = logging.INFO):
        super().__init__(interval=interval)
        self._log_level = log_level

    def do_interval(self, trainer: L.Trainer, pl_module: L.LightningModule, current_time, start_time):
        # get missing vars
        trainer_max_epochs = trainer.max_epochs if (trainer.max_epochs is not None) else float("inf")
        trainer_max_steps = trainer.max_steps if (trainer.max_steps is not None) else float("inf")

        # compute vars
        max_batches = trainer.num_training_batches  # can be inf
        max_epochs = min(trainer_max_epochs, (trainer_max_steps + max_batches - 1) // max_batches)
        max_steps = min(trainer_max_epochs * max_batches, trainer_max_steps)
        elapsed_sec = current_time - start_time

        # get vars
        global_step = trainer.global_step + 1
        epoch = trainer.current_epoch + 1
        if hasattr(trainer, "batch_idx"):
            batch = trainer.batch_idx + 1
        else:
            # warnings.warn('batch_idx missing on L.Trainer')  # TODO: re-enable this warning but only ever print once!
            batch = global_step % max_batches  # might not be int?

        # completion
        train_pct = global_step / max_steps
        train_remain_time = elapsed_sec * (1 - train_pct) / train_pct  # seconds

        # get speed -- TODO: make this a moving average?
        if global_step >= elapsed_sec:
            step_speed_str = f"{global_step / elapsed_sec:4.2f}it/s"
        else:
            step_speed_str = f"{elapsed_sec / global_step:4.2f}s/it"

        # get the metrics and format them
        info_dict = {
            k: f"{v:.4g}" if isinstance(v, (int, float)) else f"{v}" for k, v in trainer.progress_bar_metrics.items()
        }

        # sort the keys placing loss entries first
        sorted_k = sorted(info_dict.keys(), key=lambda k: ("loss" != k.lower(), "loss" not in k.lower(), k))

        # log everything
        log.log(
            level=self._log_level,
            msg=f"[{int(elapsed_sec)}s, {step_speed_str}] "
            + f"EPOCH: {epoch}/{max_epochs} - {int(global_step):0{len(str(max_steps))}d}/{max_steps} "
            + f"({int(train_pct * 100):02d}%) [rem. {int(train_remain_time)}s] "
            + f"STEP: {int(batch):{len(str(max_batches))}d}/{max_batches} ({int(batch / max_batches * 100):02d}%) "
            + f'| {" ".join(f"{k}={info_dict[k]}" for k in sorted_k)}',
        )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

# custom episodes -- base
from disent.dataset.data._episodes import BaseEpisodesData
from disent.dataset.data._episodes__custom import EpisodesPickledData
from disent.dataset.data._episodes__custom import EpisodesDownloadZippedPickledData

# raw -- groundtruth
from disent.dataset.data._groundtruth import ArrayGroundTruthData
from disent.dataset.data._groundtruth import SelfContainedHdf5GroundTruthData

# raw
from disent.dataset.data._raw import ArrayDataset
from disent.dataset.data._raw import Hdf5Dataset

# groundtruth -- base
from disent.dataset.data._groundtruth import GroundTruthData
from disent.dataset.data._groundtruth import DiskGroundTruthData
from disent.dataset.data._groundtruth import NumpyFileGroundTruthData
from disent.dataset.data._groundtruth import Hdf5GroundTruthData

# groundtruth -- impl
from disent.dataset.data._groundtruth__cars3d import Cars3dData
from disent.dataset.data._groundtruth__cars3d import Cars3d64Data  # optimized version of cars3d for 64x64 images
from disent.dataset.data._groundtruth__dsprites import DSpritesData
from disent.dataset.data._groundtruth__dsprites_imagenet import DSpritesImagenetData
from disent.dataset.data._groundtruth__mpi3d import Mpi3dData
from disent.dataset.data._groundtruth__norb import SmallNorbData
from disent.dataset.data._groundtruth__norb import SmallNorb64Data
from disent.dataset.data._groundtruth__shapes3d import Shapes3dData
from disent.dataset.data._groundtruth__sprites import SpritesData
from disent.dataset.data._groundtruth__sprites import SpritesAllData

# groundtruth -- impl synthetic
from disent.dataset.data._groundtruth__xcolumns import XColumnsData
from disent.dataset.data._groundtruth__xyobject import XYObjectData
from disent.dataset.data._groundtruth__xyobject import XYObjectShadedData
from disent.dataset.data._groundtruth__xysquares import XYSquaresData
from disent.dataset.data._groundtruth__xysquares import XYSquaresMinimalData
from disent.dataset.data._groundtruth__xysquares import XYSingleSquareData



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np

from disent.dataset.util.state_space import StateSpace

# ========================================================================= #
# TEST STATE SPACE                                                          #
# ========================================================================= #


FACTOR_SIZES = [
    [5, 4, 3, 2],
    [2, 3, 4, 5],
    [2, 3, 4],
    [1, 2, 3],
    [1, 33, 1],
    [1, 1, 1],
    [1],
]


def test_discrete_state_space_single_values():
    for factor_sizes in FACTOR_SIZES:
        states = StateSpace(factor_sizes=factor_sizes)
        # check size
        assert len(states) == np.prod(factor_sizes)
        # check single values
        for i, f in enumerate(states.factor_sizes):
            factors = states.factor_sizes - 1
            factors[:i] = 0
            idx = np.prod(states.factor_sizes[i:]) - 1
            assert states.pos_to_idx(factors) == idx
            assert np.all(states.idx_to_pos(idx) == factors)


def test_discrete_state_space_one_to_one():
    for factor_sizes in FACTOR_SIZES:
        states = StateSpace(factor_sizes=factor_sizes)
        # check that entire range of values is generated
        k = np.random.randint(1, 5)
        # chances of this failing are extremely low, but it could happen...
        pos_0 = states.sample_factors([int(100_000 ** (1 / k))] * k)
        # check random values are in the right ranges
        all_dims = tuple(range(pos_0.ndim))
        assert np.all(np.max(pos_0, axis=all_dims[:-1]) == (states.factor_sizes - 1))
        assert np.all(np.min(pos_0, axis=all_dims[:-1]) == 0)
        # check that converting between them keeps values the same
        idx_0 = states.pos_to_idx(pos_0)
        pos_1 = states.idx_to_pos(idx_0)
        idx_1 = states.pos_to_idx(pos_1)
        assert np.all(idx_0 == idx_1)
        assert np.all(pos_0 == pos_1)


def test_new_functions():
    # TODO: convert to propper tests
    s = StateSpace([2, 3, 4])

    assert np.max([s.sample_factors(size=2, f_idxs=[0, 2]) for i in range(1000)], axis=0).tolist() == [[1, 3], [1, 3]]
    assert np.max([s.sample_factors(size=(2,), f_idxs=[0, 2]) for i in range(1000)], axis=0).tolist() == [
        [1, 3],
        [1, 3],
    ]
    assert np.max([s.sample_factors(size=(2, 3), f_idxs=[0, 2]) for i in range(1000)], axis=0).tolist() == [
        [[1, 3], [1, 3], [1, 3]],
        [[1, 3], [1, 3], [1, 3]],
    ]
    assert np.max(
        [s.sample_missing_factors([[0, 0], [0, 0]], f_idxs=[0, 2]) for i in range(1000)], axis=0
    ).tolist() == [[0, 2, 0], [0, 2, 0]]

    states = StateSpace([2, 3, 4, 5])

    assert np.max(
        [states.resample_given_factors([[0, 0, 0, 0], [0, 0, 0, 0]], f_idxs=1) for i in range(1000)], axis=0
    ).tolist() == [[0, 2, 0, 0], [0, 2, 0, 0]]
    assert np.max(
        [states.resample_given_factors([[0, 0, 0, 0], [0, 0, 0, 0]], f_idxs=[1, 2]) for i in range(1000)], axis=0
    ).tolist() == [[0, 2, 3, 0], [0, 2, 3, 0]]
    assert np.max(
        [states.resample_given_factors([0, 0, 0, 0], f_idxs=[0, 3]) for i in range(1000)], axis=0
    ).tolist() == [1, 0, 0, 4]

    assert np.max(
        [states.resample_other_factors([[0, 0, 0, 0], [0, 0, 0, 0]], f_idxs=1) for i in range(1000)], axis=0
    ).tolist() == [[1, 0, 3, 4], [1, 0, 3, 4]]
    assert np.max(
        [states.resample_other_factors([[0, 0, 0, 0], [0, 0, 0, 0]], f_idxs=[1, 2]) for i in range(1000)], axis=0
    ).tolist() == [[1, 0, 0, 4], [1, 0, 0, 4]]
    assert np.max(
        [states.resample_other_factors([0, 0, 0, 0], f_idxs=[0, 3]) for i in range(1000)], axis=0
    ).tolist() == [0, 2, 3, 0]

    # make sure these work
    states.sample_missing_factors([1], f_idxs=2)
    states.sample_missing_factors([1], f_idxs=2)
    states.sample_missing_factors([[1], [3], [5]], f_idxs=2)

    # make sure these work
    states.sample_random_factor_traversal_grid(base_factors=[1, 2, 3, 4])
    states.sample_random_factor_traversal_grid(base_factors=np.array([1, 2, 3, 4]))

    # print(np.min([s.resample_radius([[0, 1, 2], [0, 0, 0]], resample_radius=1, distinct=True) for i in range(1000)], axis=0).tolist())
    # print(np.max([s.resample_radius([[0, 1, 2], [0, 0, 0]], resample_radius=1, distinct=True) for i in range(1000)], axis=0).tolist())


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2023 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import os
from pathlib import Path
from typing import Union


def tar_safe_extract_all(in_file: Union[Path, str], out_dir: Union[Path, str]):
    import tarfile

    in_file = str(in_file)
    out_dir = str(out_dir)

    def _is_safe_to_extract(tar):
        for member in tar.getmembers():
            # check inside directory
            abs_dir = os.path.abspath(out_dir)
            abs_targ = os.path.abspath(os.path.join(out_dir, member.name))
            common_prefix = os.path.commonprefix([abs_dir, abs_targ])
            # raise exception if not
            if common_prefix != abs_dir:
                raise Exception("Attempted path traversal in tar file")

    # this is unsafe tar extraction
    with tarfile.open(in_file) as f:
        _is_safe_to_extract(f)
        f.extractall(out_dir)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2022 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import numpy as np
from tqdm import tqdm

from disent.util.inout.files import AtomicSaveFile

# ========================================================================= #
# Save Numpy Files                                                          #
# ========================================================================= #


def save_dataset_array(array: np.ndarray, out_file: str, overwrite: bool = False, save_key: str = "images"):
    assert array.ndim == 4, f"invalid array shape, got: {array.shape}, must be: (N, H, W, C)"
    assert array.dtype == "uint8", f'invalid array dtype, got: {array.dtype}, must be: "uint8"'
    # save the data
    with AtomicSaveFile(out_file, overwrite=overwrite) as temp_file:
        np.savez_compressed(temp_file, **{save_key: array})


def save_resized_dataset_array(
    array: np.ndarray,
    out_file: str,
    size: int = 64,
    overwrite: bool = False,
    save_key: str = "images",
    progress: bool = True,
):
    import torchvision.transforms.functional as F_tv

    from disent.dataset.data import ArrayDataset

    # checks
    assert out_file.endswith(".npz"), f'The output file must end with the extension: ".npz", got: {repr(out_file)}'

    # Get the transform -- copied from: ToImgTensorF32 / ToImgTensorU8
    def transform(obs):
        H, W, C = obs.shape
        obs = F_tv.to_pil_image(obs)
        obs = F_tv.resize(obs, size=[size, size])
        obs = np.array(obs)
        # add removed dimension!
        if obs.ndim == 2:
            obs = obs[:, :, None]
            assert obs.shape == (size, size, C)
        return obs

    # load the converted cars3d data ?x128x128x3
    assert array.ndim == 4, f"invalid array shape, got: {array.shape}, must be: (N, H, W, C)"
    assert array.dtype == "uint8", f'invalid array dtype, got: {array.dtype}, must be: "uint8"'
    N, H, W, C = array.shape
    data = ArrayDataset(array, transform=transform)
    # save the data
    with AtomicSaveFile(out_file, overwrite=overwrite) as temp_file:
        # resize the cars3d data
        idxs = tqdm(range(N), desc="converting") if progress else range(N)
        converted = np.zeros([N, size, size, C], dtype="uint8")
        for i in idxs:
            converted[i, ...] = data[i]
        # save the data
        np.savez_compressed(temp_file, **{save_key: converted})


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
from datetime import datetime
from typing import Callable
from typing import List
from typing import NoReturn
from typing import Optional
from typing import Tuple
from typing import Union

import hydra
import lightning as L
import torch
import torch.utils.data
import wandb
from lightning import Callback
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.callbacks import ModelSummary
from lightning.pytorch.loggers import Logger
from omegaconf import DictConfig
from omegaconf import ListConfig
from omegaconf import OmegaConf

import disent.registry as R
from disent.frameworks import DisentFramework
from disent.util.lightning.callbacks import VaeMetricLoggingCallback
from disent.util.seeds import seed
from disent.util.strings import colors as c
from disent.util.strings.fmt import make_box_str
from experiment.util.hydra_data import HydraDataModule
from experiment.util.hydra_main import EXP_CONFIG_DIR
from experiment.util.hydra_main import hydra_main
from experiment.util.run_utils import safe_unset_debug_loggers
from experiment.util.run_utils import safe_unset_debug_trainer
from experiment.util.run_utils import set_debug_loggers
from experiment.util.run_utils import set_debug_trainer

log = logging.getLogger(__name__)


# ========================================================================= #
# HYDRA CONFIG HELPERS                                                      #
# ========================================================================= #


def hydra_register_disent_plugins(cfg):
    # TODO: there should be a plugin mechanism for disent?
    if cfg.experiment.plugins:
        log.info("Running experiment plugins:")
        for plugin in cfg.experiment.plugins:
            log.info(f"* registering: {plugin}")
            hydra.utils.instantiate(dict(_target_=plugin))
    else:
        log.info(
            "No experiment plugins were listed. Register these under the `experiment.plugins` in the config, which lists targets of functions."
        )


def hydra_get_accelerator_and_devices(cfg) -> Tuple[str, Optional[int]]:
    # TODO: rather specify accelerator and devices directly in config...
    #       this is redundant with the new pytorch lightning auto system.
    #       - we should also allow different accelerators like mps for apple silicon
    use_cuda = cfg.dsettings.trainer.cuda

    # check cuda values
    if use_cuda in {"try_cuda", None}:
        use_cuda = torch.cuda.is_available()
        if not use_cuda:
            log.warning("CUDA was requested, but not found on this system... CUDA has been disabled!")
    elif use_cuda:
        if not torch.cuda.is_available():
            log.error("`dsettings.trainer.cuda=True` but CUDA is not available on this machine!")
            raise RuntimeError("CUDA not available!")
    else:
        if not torch.cuda.is_available():
            log.info("CUDA is not available on this machine!")
        else:
            log.warning("CUDA is available but is not being used!")
    # get number of gpus to use
    return ("cuda", 1) if use_cuda else ("cpu", "auto")


def hydra_check_data_paths(cfg):
    prepare_data_per_node = cfg.datamodule.prepare_data_per_node
    data_root = cfg.dsettings.storage.data_root
    # check relative paths
    if not os.path.isabs(data_root):
        log.warning(
            f"A relative path was specified for dsettings.storage.data_root={repr(data_root)}."
            f" This is probably an error! Using relative paths can have unintended consequences"
            f" and performance drawbacks if the current working directory is on a shared/network drive."
            f" Hydra config also uses a new working directory for each run of the program, meaning"
            f" data will be repeatedly downloaded."
        )
        if prepare_data_per_node:
            log.error(
                f"datamodule.prepare_data_per_node={repr(prepare_data_per_node)} but dsettings.storage.data_root="
                f"{repr(data_root)} is a relative path which may be an error! Try specifying an"
                f" absolute path that is guaranteed to be unique from each node, eg. default_settings.storage.data_root=/tmp/dataset"
            )
        raise RuntimeError(f"default_settings.storage.data_root={repr(data_root)} is a relative path!")


def hydra_check_data_meta(cfg):
    # checks
    if (cfg.dataset.meta.vis_mean is None) or (cfg.dataset.meta.vis_std is None):
        log.warning(f"Dataset has no normalisation values... Are you sure this is correct?")
        log.warning(f"* dataset.meta.vis_mean: {cfg.dataset.meta.vis_mean}")
        log.warning(f"* dataset.meta.vis_std:  {cfg.dataset.meta.vis_std}")
    else:
        log.info(f"Dataset has normalisation values!")
        log.info(f"* dataset.meta.vis_mean: {cfg.dataset.meta.vis_mean}")
        log.info(f"* dataset.meta.vis_std:  {cfg.dataset.meta.vis_std}")


def hydra_make_loggers(cfg) -> List[Logger]:
    loggers = hydra.utils.instantiate(cfg.logging.loggers)
    if loggers:
        if isinstance(loggers, Logger):
            loggers = [loggers]
        for logger in loggers:
            if not isinstance(logger, Logger):
                raise TypeError(f"logger is not an instance of {Logger}, got type: {type(logger)} with value: {logger}")
        log.info(f"Initialised Loggers: {loggers}")
    else:
        loggers = []
        log.warning(f"No Logger Utilised!")
    return loggers


def hydra_get_callbacks(cfg) -> list:
    callbacks = []
    # add all callbacks
    for name, item in cfg.callbacks.items():
        # custom callback handling vs instantiation
        callback = hydra.utils.instantiate(item)
        assert isinstance(
            callback, Callback
        ), f"instantiated callback is not an instance of {Callback}, got: {callback}"
        # add to callbacks list
        log.info(f"made callback: {name} ({item._target_})")
        callbacks.append(callback)
    return callbacks


def hydra_get_checkpoint_callbacks(cfg) -> list:
    callbacks = []
    if cfg.settings.checkpoint.save_checkpoint:
        hydra_ckp_dir = os.path.join(os.getcwd(), "checkpoints")
        hydra_checkpoint = ModelCheckpoint(dirpath=hydra_ckp_dir, verbose=True, save_last=True)
        callbacks.append(hydra_checkpoint)
        if cfg.logging.wandb.enabled:
            wandb_ckp_dir = os.path.join(wandb.run.dir, "checkpoints")
            wandb_checkpoint = ModelCheckpoint(dirpath=wandb_ckp_dir, save_last=True)
            callbacks.append(wandb_checkpoint)
    return callbacks


def hydra_get_metric_callbacks(cfg) -> list:
    # TODO: simplify this, make better use of the config!
    callbacks = []
    # set default values used later
    default_every_n_steps = cfg.metrics.default_every_n_steps
    default_on_final = cfg.metrics.default_on_final
    default_on_train = cfg.metrics.default_on_train
    default_begin_first_step = cfg.metrics.default_begin_first_step
    # get metrics
    metric_list = cfg.metrics.metric_list
    assert isinstance(metric_list, (list, ListConfig)), f"`metrics.metric_list` is not a list, got: {type(metric_list)}"
    # get metrics
    for metric in metric_list:
        assert isinstance(
            metric, (dict, DictConfig)
        ), f"entry in metric list is not a dictionary, got type: {type(metric)} or value: {repr(metric)}"
        # fix the values
        if isinstance(metric, str):
            metric = {metric: {}}
        ((name, settings),) = metric.items()
        # check values
        assert isinstance(
            metric, (dict, DictConfig)
        ), f"settings for entry in metric list is not a dictionary, got type: {type(settings)} or value: {repr(settings)}"
        # make metrics
        train_metric = [R.METRICS[name].compute_fast] if settings.get("on_train", default_on_train) else None
        final_metric = [R.METRICS[name].compute] if settings.get("on_final", default_on_final) else None
        # add the metric callback
        if final_metric or train_metric:
            callbacks.append(
                VaeMetricLoggingCallback(
                    step_end_metrics=train_metric,
                    train_end_metrics=final_metric,
                    every_n_steps=settings.get("every_n_steps", default_every_n_steps),
                    begin_first_step=settings.get("begin_first_step", default_begin_first_step),
                )
            )
    return callbacks


def hydra_create_framework(
    cfg, gpu_batch_augment: Optional[Callable[[torch.Tensor], torch.Tensor]] = None
) -> DisentFramework:
    # create framework
    assert str.endswith(
        cfg.framework.cfg["_target_"], ".cfg"
    ), f'`cfg.framework.cfg._target_` does not end with ".cfg", got: {repr(cfg.framework.cfg["_target_"])}'
    framework_cls = hydra.utils.get_class(cfg.framework.cfg["_target_"][: -len(".cfg")])
    framework: DisentFramework = framework_cls(
        model=hydra.utils.instantiate(cfg.model.model_cls),
        cfg=hydra.utils.instantiate(
            cfg.framework.cfg, _convert_="all"
        ),  # DisentConfigurable -- convert all OmegaConf objects to python equivalents, eg. DictConfig -> dict
        batch_augment=gpu_batch_augment,
    )

    # check if some cfg variables were not overridden
    missing_keys = sorted(set(framework.cfg.get_keys()) - (set(cfg.framework.cfg.keys())))
    if missing_keys:
        log.warning(f"{c.RED}Framework {repr(cfg.framework.name)} is missing config keys for:{c.RST}")
        for k in missing_keys:
            log.warning(f"{c.RED}{repr(k)}{c.RST}")

    # register schedules to the framework
    schedule_items = cfg.schedule.schedule_items
    assert isinstance(
        schedule_items, (dict, DictConfig)
    ), f"`schedule.schedule_items` must be a dictionary, got type: {type(schedule_items)} with value: {repr(schedule_items)}"
    if schedule_items:
        log.info(f"Registering Schedules:")
        for target, schedule in schedule_items.items():
            framework.register_schedule(target, hydra.utils.instantiate(schedule), logging=True)

    return framework


def hydra_make_datamodule(cfg):
    return HydraDataModule(
        data=cfg.dataset.data,  # from: dataset
        transform=cfg.dataset.transform,  # from: dataset
        augment=cfg.augment.augment_cls,  # from: augment
        sampler=cfg.sampling._sampler_.sampler_cls,  # from: sampling
        # from: run_location
        using_cuda=cfg.dsettings.trainer.cuda,
        dataloader_kwargs=cfg.datamodule.dataloader,
        augment_on_gpu=cfg.datamodule.gpu_augment,
        prepare_data_per_node=cfg.datamodule.prepare_data_per_node,
        # from: framework.meta
        return_indices=cfg.framework.meta.get("requires_indices", False),
        return_factors=cfg.framework.meta.get("requires_factors", False),
    )


# ========================================================================= #
# ACTIONS                                                                   #
# ========================================================================= #


def action_prepare_data(cfg: DictConfig):
    # get the time the run started
    time_string = datetime.today().strftime("%Y-%m-%d--%H-%M-%S")
    log.info(f"Starting run at time: {time_string}")
    # deterministic seed
    seed(cfg.settings.job.seed)
    # register plugins
    hydra_register_disent_plugins(cfg)
    # print useful info
    log.info(f"Current working directory : {os.getcwd()}")
    log.info(f"Orig working directory    : {hydra.utils.get_original_cwd()}")
    # check data preparation
    hydra_check_data_paths(cfg)
    hydra_check_data_meta(cfg)
    # print the config
    log.info(f'Dataset Config Is:\n{make_box_str(OmegaConf.to_yaml({"dataset": cfg.dataset}))}')
    # prepare data
    datamodule = hydra_make_datamodule(cfg)
    datamodule.prepare_data()


def action_train(cfg: DictConfig):
    # get the time the run started
    time_string = datetime.today().strftime("%Y-%m-%d--%H-%M-%S")
    log.info(f"Starting run at time: {time_string}")

    # -~-~-~-~-~-~-~-~-~-~-~-~- #
    # cleanup from old runs:
    # -~-~-~-~-~-~-~-~-~-~-~-~- #

    try:
        safe_unset_debug_trainer()
        safe_unset_debug_loggers()
        wandb.finish()
    except:
        pass

    # -~-~-~-~-~-~-~-~-~-~-~-~- #
    # SETUP
    # -~-~-~-~-~-~-~-~-~-~-~-~- #

    # create trainer loggers & callbacks & initialise error messages
    loggers = set_debug_loggers(hydra_make_loggers(cfg))

    # deterministic seed
    seed(cfg.settings.job.seed)
    # register plugins
    hydra_register_disent_plugins(cfg)
    # print useful info
    log.info(f"Current working directory : {os.getcwd()}")
    log.info(f"Orig working directory    : {hydra.utils.get_original_cwd()}")
    # checks
    accelerator, devices = hydra_get_accelerator_and_devices(cfg)
    hydra_check_data_paths(cfg)
    hydra_check_data_meta(cfg)

    # -~-~-~-~-~-~-~-~-~-~-~-~- #
    # INITIALISE
    # -~-~-~-~-~-~-~-~-~-~-~-~- #

    # HYDRA MODULES
    datamodule = hydra_make_datamodule(cfg)
    framework = hydra_create_framework(cfg, gpu_batch_augment=datamodule.gpu_batch_augment)

    # TRAINER
    # - trainer: callbacks
    trainer_callbacks = [
        *hydra_get_callbacks(cfg),
        *hydra_get_checkpoint_callbacks(cfg),
        *hydra_get_metric_callbacks(cfg),
        ModelSummary(max_depth=2),  # override default ModelSummary set by trainer
    ]

    # - trainer: default kwargs
    trainer_default_kwargs = dict(
        detect_anomaly=False,  # this should only be enabled for debugging torch and finding NaN values, slows down execution, not by much though?
        enable_checkpointing=cfg.settings.checkpoint.save_checkpoint,
    )

    # - trainer: init
    trainer = set_debug_trainer(
        L.Trainer(
            # cannot override these
            logger=loggers,
            accelerator=accelerator,
            devices=devices,
            callbacks=trainer_callbacks,
            # additional kwargs from the config, overrides the defaults
            **{**trainer_default_kwargs, **cfg.trainer},
        )
    )

    # -~-~-~-~-~-~-~-~-~-~-~-~- #
    # DEBUG
    # -~-~-~-~-~-~-~-~-~-~-~-~- #

    # get config sections
    print_cfg, boxed_pop = dict(cfg), lambda *keys: make_box_str(
        OmegaConf.to_yaml({k: print_cfg.pop(k) for k in keys} if keys else print_cfg)
    )
    cfg_str_exp = boxed_pop("action", "experiment")
    cfg_str_logging = boxed_pop("logging", "callbacks", "metrics")
    cfg_str_dataset = boxed_pop("dataset", "datamodule", "sampling", "augment")
    cfg_str_system = boxed_pop("framework", "model", "schedule")
    cfg_str_settings = boxed_pop("dsettings", "settings")
    cfg_str_other = boxed_pop()
    # print config sections
    log.info(
        f"Final Config For Action: {cfg.action}\n\nEXPERIMENT:{cfg_str_exp}\nLOGGING:{cfg_str_logging}\nDATASET:{cfg_str_dataset}\nSYSTEM:{cfg_str_system}\nTRAINER:{cfg_str_other}\nSETTINGS:{cfg_str_settings}"
    )

    # -~-~-~-~-~-~-~-~-~-~-~-~- #
    # BEGIN TRAINING
    # -~-~-~-~-~-~-~-~-~-~-~-~- #

    # save hparams
    framework.hparams.update(cfg)

    # TODO: is this a pytorch lightning bug? The trainer should automatically save these if hparams is set?
    if trainer.loggers:
        for logger in trainer.loggers:
            logger.log_hyperparams(framework.hparams)

    # fit the model
    # -- if an error/signal occurs while pytorch lightning is
    #    initialising the training process we cannot capture it!
    trainer.fit(framework, datamodule=datamodule)

    # -~-~-~-~-~-~-~-~-~-~-~-~- #
    # cleanup this run
    # -~-~-~-~-~-~-~-~-~-~-~-~- #

    try:
        wandb.finish()
    except:
        pass


# available actions
ACTIONS = {
    "prepare_data": action_prepare_data,
    "train": action_train,
    "skip": lambda *args, **kwargs: None,
}


def run_action(cfg: DictConfig):
    action_key = cfg.action
    # get the action
    if action_key not in ACTIONS:
        raise KeyError(f"The given action: {repr(action_key)} is invalid, must be one of: {sorted(ACTIONS.keys())}")
    action = ACTIONS[action_key]
    # run the action
    action(cfg)


# ========================================================================= #
# MAIN                                                                      #
# ========================================================================= #


def hydra_experiment(
    callback: Callable[[DictConfig], NoReturn] = run_action,
    config_name: str = "config",
    # config search path
    search_dir_main: str = EXP_CONFIG_DIR,
    search_dirs_prepend: Optional[Union[str, List[str]]] = None,
    search_dirs_append: Optional[Union[str, List[str]]] = None,
    # logging
    log_level: Optional[int] = logging.INFO,
    log_exc_info_callback: bool = True,
    log_exc_info_hydra: bool = False,
):
    """
    Same as `hydra_main`, but with defaults to use
    with this experiment directory!
    """
    hydra_main(
        callback=callback,
        config_name=config_name,
        search_dir_main=search_dir_main,
        search_dirs_prepend=search_dirs_prepend,
        search_dirs_append=search_dirs_append,
        log_level=log_level,
        log_exc_info_callback=log_exc_info_callback,
        log_exc_info_hydra=log_exc_info_hydra,
    )


# launch the action
if __name__ == "__main__":
    hydra_experiment()


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



# coding=utf-8
# Copyright 2018 The DisentanglementLib Authors.  All rights reserved.
# https://github.com/google-research/disentanglement_lib
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# CHANGES:
# - converted from tensorflow to pytorch
# - removed gin config
# - uses disent objects and classes
# - renamed functions

"""
Implementation of the disentanglement metric from the FactorVAE paper.
Based on "Disentangling by Factorising" (https://arxiv.org/abs/1802.05983).
"""

import logging

import numpy as np
from tqdm import tqdm

from disent.dataset import DisentDataset
from disent.metrics import utils
from disent.metrics.utils import make_metric
from disent.util import to_numpy

log = logging.getLogger(__name__)


# ========================================================================= #
# factor_vae                                                                #
# ========================================================================= #


@make_metric(
    "factor_vae", fast_kwargs=dict(num_train=700, num_eval=350, num_variance_estimate=1000)
)  # may not be accurate, but it just takes waay too long otherwise 20+ seconds
def metric_factor_vae(
    dataset: DisentDataset,
    representation_function: callable,
    batch_size: int = 64,
    num_train: int = 10000,
    num_eval: int = 5000,
    num_variance_estimate: int = 10000,
    show_progress=False,
):
    """
    Computes the FactorVAE disentanglement metric.

    Algorithm Description (Excerpt from paper):
    =====================

    1. Choose a factor k
    2. generate data with this factor fixed but all other factors varying randomly;
    3. obtain their representations;
    4. normalise each dimension by its empirical standard deviation over the full data
      (or a large enough random subset);
    5. take the empirical variance in each dimension of these normalised representations.
    6. Then the index of the dimension with the lowest variance and the target index k
       provide one training input/output example for the classifier (see bottom of Figure 2).

    # ---------------------------------------------------------------------- #
    | Thus if the representation is perfectly disentangled, the empirical    |
    | variance in the dimension corresponding to the fixed factor will be 0. |
    # ---------------------------------------------------------------------- #

    - We normalise the representations (above) so that the arg min is invariant to rescaling
      of the representations in each dimension.

    - Since both inputs and outputs lie in a discrete space, the optimal
         classifier is the majority-vote classifier (see Appendix B for details),
         and the metric is the error rate of the classifier.

    The resulting classifier is a deterministic function of the training data, hence there are no optimisation hyperparameters to tune.
    We also believe that this metric is conceptually simpler and more natural than the previous one.
    Most importantly, it circumvents the failure mode of the earlier metric, since the classifier needs to see the lowest variance in a latent dimension for a given factor to classify it correctly

    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and
        outputs a dim_representation sized representation for each observation.
      batch_size: Number of points to be used to compute the training_sample.
      num_train: Number of points used for training.
      num_eval: Number of points used for evaluation.
      num_variance_estimate: Number of points used to estimate global variances.
      show_progress: If a tqdm progress bar should be shown
    Returns:
      Dictionary with scores:
        train_accuracy: Accuracy on training set.
        eval_accuracy: Accuracy on evaluation set.
    """

    log.debug("Computing global variances to standardise.")
    global_variances = _compute_variances(dataset, representation_function, num_variance_estimate)
    active_dims = _prune_dims(global_variances)

    if not active_dims.any():
        return {"factor_vae.train_accuracy": 0.0, "factor_vae.eval_accuracy": 0.0, "factor_vae.num_active_dims": 0}

    log.debug("Generating training set.")
    training_votes = _generate_training_batch(
        dataset,
        representation_function,
        batch_size,
        num_train,
        global_variances,
        active_dims,
        show_progress=show_progress,
    )
    classifier = np.argmax(training_votes, axis=0)
    other_index = np.arange(training_votes.shape[1])

    # Evaluate training set accuracy
    train_accuracy = np.sum(training_votes[classifier, other_index]) * 1.0 / np.sum(training_votes)

    log.debug("Generating evaluation set.")
    eval_votes = _generate_training_batch(
        dataset,
        representation_function,
        batch_size,
        num_eval,
        global_variances,
        active_dims,
        show_progress=show_progress,
    )

    # Evaluate evaluation set accuracy
    eval_accuracy = np.sum(eval_votes[classifier, other_index]) * 1.0 / np.sum(eval_votes)

    return {
        "factor_vae.train_accuracy": train_accuracy,  # "z-min variance" -- Measuring Disentanglement: A Review of Metrics
        "factor_vae.eval_accuracy": eval_accuracy,  # "z-min variance" -- Measuring Disentanglement: A Review of Metrics
        "factor_vae.num_active_dims": len(active_dims),
    }


def _prune_dims(variances, threshold=0.0):
    """Mask for dimensions collapsed to the prior."""
    scale_z = np.sqrt(variances)
    return scale_z >= threshold


def _compute_variances(
    dataset: DisentDataset, representation_function: callable, batch_size: int, eval_batch_size: int = 64
):
    """Computes the variance for each dimension of the representation.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observation as input and outputs a representation.
      batch_size: Number of points to be used to compute the variances.
      eval_batch_size: Batch size used to eval representation.
    Returns:
      Vector with the variance of each dimension.
    """
    observations = dataset.dataset_sample_batch(batch_size, mode="input")
    representations = to_numpy(utils.obtain_representation(observations, representation_function, eval_batch_size))
    representations = np.transpose(representations)
    assert representations.shape[0] == batch_size
    return np.var(representations, axis=0, ddof=1)


def _generate_training_sample(
    dataset: DisentDataset,
    representation_function: callable,
    batch_size: int,
    global_variances: np.ndarray,
    active_dims: list,
) -> (int, int):
    """Sample a single training sample based on a mini-batch of ground-truth data.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observation as input and
        outputs a representation.
      batch_size: Number of points to be used to compute the training_sample.
      global_variances: Numpy vector with variances for all dimensions of representation.
      active_dims: Indexes of active dimensions.
    Returns:
      factor_index: Index of factor coordinate to be used.
      argmin: Index of representation coordinate with the least variance.
    """
    # Select random coordinate to keep fixed.
    factor_index = np.random.randint(dataset.gt_data.num_factors)
    # Sample two mini batches of latent variables.
    factors = dataset.gt_data.sample_factors(batch_size)
    # Fix the selected factor across mini-batch.
    factors[:, factor_index] = factors[0, factor_index]
    # Obtain the observations.
    observations = dataset.dataset_batch_from_factors(factors, mode="input")
    representations = to_numpy(representation_function(observations))
    local_variances = np.var(representations, axis=0, ddof=1)
    argmin = np.argmin(local_variances[active_dims] / global_variances[active_dims])
    return factor_index, argmin


def _generate_training_batch(
    dataset: DisentDataset,
    representation_function: callable,
    batch_size: int,
    num_points: int,
    global_variances: np.ndarray,
    active_dims: list,
    show_progress=False,
):
    """Sample a set of training samples based on a batch of ground-truth data.
    Args:
      dataset: DisentDataset to be sampled from.
      representation_function: Function that takes observations as input and outputs a dim_representation sized representation for each observation.
      batch_size: Number of points to be used to compute the training_sample.
      num_points: Number of points to be sampled for training set.
      global_variances: Numpy vector with variances for all dimensions of representation.
      active_dims: Indexes of active dimensions.
    Returns:
      (num_factors, dim_representation)-sized numpy array with votes.
    """
    votes = np.zeros((dataset.gt_data.num_factors, global_variances.shape[0]), dtype=np.int64)
    for _ in tqdm(range(num_points), disable=(not show_progress)):
        factor_index, argmin = _generate_training_sample(
            dataset, representation_function, batch_size, global_variances, active_dims
        )
        votes[factor_index, argmin] += 1
    return votes


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
from typing import Dict
from typing import Optional
from typing import Sequence
from typing import Union

import numpy as np
from tqdm import tqdm

from disent.dataset.data import DiskGroundTruthData
from disent.dataset.data import Hdf5GroundTruthData
from disent.dataset.data._groundtruth import NumpyFileGroundTruthData
from disent.dataset.util.datafile import DataFile
from disent.dataset.util.datafile import DataFileHashed
from disent.dataset.util.datafile import DataFileHashedDl
from disent.util.inout.paths import modify_ext

log = logging.getLogger(__name__)

# ========================================================================= #
# mpi3d - conversion                                                        #
# ========================================================================= #


def resave_mpi3d_array(in_npz_path: str, out_h5_path: str, overwrite: bool = False):
    from disent.dataset.util.formats.hdf5 import H5Builder
    from disent.util.profiling import Timer

    # load the array
    with Timer("loading images into memory"):
        imgs = np.load(in_npz_path)["images"]
        assert imgs.dtype == "uint8"
        assert imgs.shape == (1036800, 64, 64, 3)
    # resave the array as hdf5
    with Timer("resaving images"):
        with H5Builder(out_h5_path, mode="atomic_w" if overwrite else "atomic_x") as builder:
            builder.add_dataset_from_array(name="images", array=imgs, show_progress=True)


class DataFileMpi3dResaved(DataFileHashed):
    def __init__(
        self,
        mpi3d_datafile: DataFileHashedDl,
        # # - convert file name
        out_hash: Optional[Union[str, Dict[str, str]]],
        out_name: Optional[str] = None,
        # # - hash settings
        hash_type: str = "md5",
        hash_mode: str = "fast",
    ):
        self._mpi3d_datafile = mpi3d_datafile
        super().__init__(
            file_name=modify_ext(self._mpi3d_datafile.out_name, "h5") if (out_name is None) else out_name,
            file_hash=out_hash,
            hash_type=hash_type,
            hash_mode=hash_mode,
        )

    def _prepare(self, out_dir: str, out_file: str):
        log.debug("Preparing Orig Mpi3d Data:")
        mpi3d_path = self._mpi3d_datafile.prepare(out_dir)
        log.debug("Generating hdf5 Mpi3d Data:")
        resave_mpi3d_array(in_npz_path=mpi3d_path, out_h5_path=out_file, overwrite=True)

    # match the functionality of DataFileHashedDlH5
    dataset_name = "images"


# ========================================================================= #
# mpi3d                                                                     #
# ========================================================================= #


class _Mpi3dMixin:
    # override

    name = "mpi3d"
    factor_names = (
        "object_color",
        "object_shape",
        "object_size",
        "camera_height",
        "background_color",
        "first_dof",
        "second_dof",
    )
    factor_sizes = (6, 6, 2, 3, 3, 40, 40)  # TOTAL: 1 036 800
    img_shape = (64, 64, 3)

    @property
    def datafile(self) -> DataFile:
        assert (
            self.subset in self.MPI3D_DATAFILES
        ), f"Invalid MPI3D subset: {repr(self.subset)} must be one of: {set(self.MPI3D_DATAFILES.keys())}"
        return self.MPI3D_DATAFILES[self.subset]

    # not implemented
    _subset: str

    @property
    def MPI3D_DATAFILES(self) -> Dict[str, DataFile]:
        raise NotImplementedError

    @property
    def subset(self) -> str:
        return self._subset


class Mpi3dNumpyData(_Mpi3dMixin, NumpyFileGroundTruthData):
    """
    MPI3D Dataset
    - https://github.com/rr-learning/disentanglement_dataset

    reference implementation: https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/data/ground_truth/mpi3d.py
    """

    # override NumpyFileGroundTruthData
    data_key = "images"

    # override _Mpi3dMixin
    MPI3D_DATAFILES = {
        "toy": DataFileHashedDl(
            uri="https://storage.googleapis.com/disentanglement_dataset/Final_Dataset/mpi3d_toy.npz",
            uri_hash={"fast": "146138e36ff495e77ceacdc8cf14c37e", "full": "55889cb7c7dfc655d6e0277beee88868"},
        ),
        "realistic": DataFileHashedDl(
            uri="https://storage.googleapis.com/disentanglement_dataset/Final_Dataset/mpi3d_realistic.npz",
            uri_hash={"fast": "96c8ff1155dd61f79d3493edef9f19e9", "full": "59a6225b88b635365f70c91b3e52f70f"},
        ),
        "real": DataFileHashedDl(
            uri="https://storage.googleapis.com/disentanglement_dataset/Final_Dataset/mpi3d_real.npz",
            uri_hash={"fast": "e2941bba6f4a2b130edc5f364637b39e", "full": "0f33f609918fb5c97996692f91129802"},
        ),
    }

    def __init__(
        self, data_root: Optional[str] = None, prepare: bool = False, subset: str = "realistic", transform=None
    ):
        self._subset = subset
        log.warning("[WARNING]: mpi3d files are extremely large (over 11GB), you are trying to load these into memory.")
        super().__init__(data_root=data_root, prepare=prepare, transform=transform)


class Mpi3dHdf5Data(_Mpi3dMixin, Hdf5GroundTruthData):
    """
    MPI3D Dataset
    - https://github.com/rr-learning/disentanglement_dataset

    reference implementation: https://github.com/google-research/disentanglement_lib/blob/master/disentanglement_lib/data/ground_truth/mpi3d.py
    """

    MPI3D_DATAFILES = {
        "toy": DataFileMpi3dResaved(
            mpi3d_datafile=Mpi3dNumpyData.MPI3D_DATAFILES["toy"],
            out_hash={"fast": "32af615e306d336449a83c53ee897c3f", "full": "???"},
        ),
        "realistic": DataFileMpi3dResaved(
            mpi3d_datafile=Mpi3dNumpyData.MPI3D_DATAFILES["realistic"],
            out_hash={"fast": "c6bfa8ecec2549dbf7f0c29f853a6beb", "full": "???"},
        ),
        "real": DataFileMpi3dResaved(
            mpi3d_datafile=Mpi3dNumpyData.MPI3D_DATAFILES["real"],
            out_hash={"fast": "81b9c54c75e921a2465971911b096ac3", "full": "???"},
        ),
    }

    def __init__(
        self,
        data_root: Optional[str] = None,
        prepare: bool = False,
        subset: str = "realistic",
        in_memory: bool = False,
        transform=None,
    ):
        self._subset = subset
        if in_memory:
            log.warning(
                "[WARNING]: mpi3d files are extremely large (over 11GB), you are trying to load these into memory."
            )
        super().__init__(data_root=data_root, prepare=prepare, in_memory=in_memory, transform=transform)


class Mpi3dData(DiskGroundTruthData):
    # override
    name = _Mpi3dMixin.name
    factor_names = _Mpi3dMixin.factor_names
    factor_sizes = _Mpi3dMixin.factor_sizes
    img_shape = _Mpi3dMixin.img_shape

    def __init__(
        self,
        data_root: Optional[str] = None,
        prepare: bool = False,
        subset: str = "realistic",
        in_memory: bool = False,
        transform=None,
    ):
        if in_memory:
            self._wrapped_mpi3d = Mpi3dNumpyData(data_root=data_root, prepare=False, subset=subset, transform=None)
        else:
            self._wrapped_mpi3d = Mpi3dHdf5Data(
                data_root=data_root, prepare=False, subset=subset, transform=None, in_memory=False
            )
        # initialize
        super().__init__(data_root=data_root, prepare=prepare, transform=transform)

    @property
    def datafiles(self) -> Sequence[DataFile]:
        return self._wrapped_mpi3d.datafiles

    def _get_observation(self, idx):
        return self._wrapped_mpi3d._get_observation(idx)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)

    def main():
        from disent.util.profiling import Timer

        # NAME:     | both-disk | both-mem | numpy-mem | hdf5-disk | hdf5-mem
        # ----------+-----------+----------+-----------+-----------+---------
        # real      | 908.512µs | 10.905s  | 10.901s   | 753.156µs | 1m:11s
        # realistic | 12.397ms  | 11.039s  | 11.024s   | 764.293µs | 54.281s
        # toy       | 12.770ms  | 10.992s  | 10.984s   | 781.021µs | 38.165s

        for subset in ["real", "realistic", "toy"]:
            print("=" * 100)
            print(subset)
            print("=" * 100)

            with Timer("both-disk"):
                both_dist = Mpi3dData(subset=subset, prepare=True, in_memory=False)
            with Timer("hdf5-disk"):
                hdf5_disk = Mpi3dHdf5Data(subset=subset, prepare=True, in_memory=False)
            with Timer("both-mem"):
                both_mem = Mpi3dData(subset=subset, prepare=True, in_memory=True)
            with Timer("numpy-mem"):
                numpy_mem = Mpi3dNumpyData(subset=subset, prepare=True)
            with Timer("hdf5-mem"):
                hdf5_mem = Mpi3dHdf5Data(subset=subset, prepare=True, in_memory=True)  # this is slow!

            assert len(both_dist) == len(hdf5_disk)
            assert len(hdf5_disk) == len(both_mem)
            assert len(both_mem) == len(numpy_mem)
            assert len(numpy_mem) == len(hdf5_mem)

            # check equivalence
            for bd, hd, bm, nm, hm in tqdm(zip(both_dist, hdf5_disk, both_mem, numpy_mem, hdf5_mem)):
                assert np.allclose(bd, hd)
                assert np.allclose(hd, bm)
                assert np.allclose(bm, nm)
                assert np.allclose(nm, hm)

    main()



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from typing import Sequence
from typing import Union
from typing import final

import torch
import torch.nn.functional as F

import disent.registry as R
from disent.dataset.transform import FftKernel
from disent.frameworks.helper.util import compute_ave_loss
from disent.nn.loss.reduction import batch_loss_reduction
from disent.nn.loss.reduction import loss_reduction
from disent.nn.modules import DisentModule
from disent.util.deprecate import deprecated

# ========================================================================= #
# Reconstruction Loss Base                                                  #
# ========================================================================= #


class ReconLossHandler(DisentModule):
    def __init__(self, reduction: str = "mean"):
        super().__init__()
        self._reduction = reduction

    def forward(self, *args, **kwargs):
        raise RuntimeError(f"Cannot call forward() on {self.__class__.__name__}")

    def activate(self, x_partial: torch.Tensor):
        """
        The final activation of the model.
        - Never use this in a training loop.
        """
        raise NotImplementedError

    def activate_all(self, xs_partial: Sequence[torch.Tensor]):
        return [self.activate(x_partial) for x_partial in xs_partial]

    @final
    def compute_loss(self, x_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        """
        Takes in activated tensors
        :return: The computed reduced loss
        """
        assert x_recon.shape == x_targ.shape, f"x_recon.shape={x_recon.shape} x_targ.shape={x_targ.shape}"
        batch_loss = self.compute_unreduced_loss(x_recon, x_targ)
        loss = loss_reduction(batch_loss, reduction=self._reduction)
        return loss

    @final
    def compute_loss_from_partial(self, x_partial_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        """
        Takes in an **unactivated** tensor from the model
        as well as an original target from the dataset.
        :return: The computed reduced loss
        """
        assert (
            x_partial_recon.shape == x_targ.shape
        ), f"x_partial_recon.shape={x_partial_recon.shape} x_targ.shape={x_targ.shape}"
        batch_loss = self.compute_unreduced_loss_from_partial(x_partial_recon, x_targ)
        loss = loss_reduction(batch_loss, reduction=self._reduction)
        return loss

    @final
    def compute_ave_loss(self, xs_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]) -> torch.Tensor:
        """
        Compute the average over losses computed from corresponding tensor pairs in the sequence.
        """
        return compute_ave_loss(self.compute_loss, xs_recon, xs_targ)

    @final
    def compute_ave_loss_from_partial(
        self, xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> torch.Tensor:
        """
        Compute the average over losses computed from corresponding tensor pairs in the sequence.
        """
        return compute_ave_loss(self.compute_loss_from_partial, xs_partial_recon, xs_targ)

    def compute_unreduced_loss(self, x_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        """
        Takes in activated tensors
        Compute the loss without applying a reduction, the loss
        tensor should be the same shapes as the input tensors
        :return: The computed unreduced loss
        """
        raise NotImplementedError

    def compute_unreduced_loss_from_partial(self, x_partial_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        """
        Takes in an **unactivated** tensor from the model
        Compute the loss without applying a reduction, the loss
        tensor should be the same shapes as the input tensors
        :return: The computed unreduced loss
        """
        raise NotImplementedError

    def _pairwise_reduce(self, unreduced_loss: torch.Tensor):
        assert self._reduction in ("mean", "sum"), f'pairwise losses only support "mean" and "sum" reduction modes.'
        return batch_loss_reduction(unreduced_loss, reduction_dtype=None, reduction=self._reduction)

    def compute_pairwise_loss(self, x_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        return self._pairwise_reduce(self.compute_unreduced_loss(x_recon, x_targ))

    def compute_pairwise_loss_from_partial(self, x_partial_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        return self._pairwise_reduce(self.compute_unreduced_loss_from_partial(x_partial_recon, x_targ))


# ========================================================================= #
# Reconstruction Losses                                                     #
# ========================================================================= #


class ReconLossHandlerMse(ReconLossHandler):
    """
    MSE loss should be used with continuous targets between [0, 1].
    - using BCE for such targets is a prevalent error in VAE research.
    """

    def activate(self, x_partial: torch.Tensor) -> torch.Tensor:
        # mse requires no final activation
        return x_partial

    def compute_unreduced_loss(self, x_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        return F.mse_loss(x_recon, x_targ, reduction="none")

    def compute_unreduced_loss_from_partial(self, x_partial_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        return self.compute_unreduced_loss(self.activate(x_partial_recon), x_targ)


class ReconLossHandlerMae(ReconLossHandlerMse):
    """
    MAE loss should be used with continuous targets between [0, 1].
    """

    def compute_unreduced_loss(self, x_recon, x_targ):
        return torch.abs(x_recon - x_targ)


class ReconLossHandlerBce(ReconLossHandler):
    """
    BCE loss should only be used with binary targets {0, 1}.
    - ignoring this and not using MSE is a prevalent error in VAE research.
    """

    def activate(self, x_partial: torch.Tensor):
        # we allow the model output x to generally be in the range [-1, 1] and scale
        # it to the range [0, 1] here to match the targets.
        return torch.sigmoid(x_partial)

    def compute_unreduced_loss(self, x_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        warnings.warn("binary cross entropy not computed over logits is inaccurate!")
        return F.binary_cross_entropy(x_recon, x_targ, reduction="none")

    def compute_unreduced_loss_from_partial(self, x_partial_recon, x_targ):
        """
        Computes the Bernoulli loss for the sigmoid activation function
        REFERENCE:
            https://github.com/google-research/disentanglement_lib/blob/76f41e39cdeff8517f7fba9d57b09f35703efca9/disentanglement_lib/methods/shared/losses.py
            - the same when reduction=='mean_sum' for super().training_compute_loss()
        REFERENCE ALT:
            https://github.com/YannDubs/disentangling-vae/blob/master/disvae/models/losses.py
        """
        return F.binary_cross_entropy_with_logits(x_partial_recon, x_targ, reduction="none")


# ========================================================================= #
# Reconstruction Distributions                                              #
# ========================================================================= #


class ReconLossHandlerBernoulli(ReconLossHandlerBce):
    def compute_unreduced_loss(self, x_recon, x_targ):
        # This is exactly the same as the BCE version, but more 'correct'.
        warnings.warn("bernoulli not computed over logits might be inaccurate!")
        return -torch.distributions.Bernoulli(probs=x_recon).log_prob(x_targ)

    def compute_unreduced_loss_from_partial(self, x_partial_recon, x_targ):
        # This is exactly the same as the BCE version, but more 'correct'.
        return -torch.distributions.Bernoulli(logits=x_partial_recon).log_prob(x_targ)


class ReconLossHandlerContinuousBernoulli(ReconLossHandlerBce):
    """
    The continuous Bernoulli: fixing a pervasive error in variational autoencoders
    - Loaiza-Ganem G and Cunningham JP, NeurIPS 2019.
    - https://arxiv.org/abs/1907.06845
    """

    def compute_unreduced_loss(self, x_recon, x_targ):
        warnings.warn("Using continuous bernoulli distribution for reconstruction loss. This is not yet recommended!")
        warnings.warn("continuous bernoulli not computed over logits might be inaccurate!")
        # I think there is something wrong with this...
        return -torch.distributions.ContinuousBernoulli(probs=x_recon, lims=(0.49, 0.51)).log_prob(x_targ)

    def compute_unreduced_loss_from_partial(self, x_partial_recon, x_targ):
        warnings.warn("Using continuous bernoulli distribution for reconstruction loss. This is not yet recommended!")
        # I think there is something wrong with this...
        return -torch.distributions.ContinuousBernoulli(logits=x_partial_recon, lims=(0.49, 0.51)).log_prob(x_targ)


class ReconLossHandlerNormal(ReconLossHandlerMse):
    def compute_unreduced_loss(self, x_recon, x_targ):
        # this is almost the same as MSE, but scaled with a tiny offset
        # A value for scale should actually be passed...
        warnings.warn("Using normal distribution for reconstruction loss. This is not yet recommended!")
        return -torch.distributions.Normal(x_recon, 1.0).log_prob(x_targ)


# ========================================================================= #
# Augmented Losses                                                          #
# ========================================================================= #


_NO_ARG = object()


class AugmentedReconLossHandler(ReconLossHandler):
    def __init__(
        self,
        recon_loss_handler: ReconLossHandler,
        kernel: Union[str, torch.Tensor],
        wrap_weight: float = 1.0,
        aug_weight: float = 1.0,
        normalize_mode: str = _NO_ARG,
    ):
        super().__init__(reduction=recon_loss_handler._reduction)
        # save variables
        self._recon_loss_handler = recon_loss_handler
        # must be a recon loss handler, but cannot nest augmented handlers
        assert isinstance(recon_loss_handler, ReconLossHandler)
        assert not isinstance(recon_loss_handler, AugmentedReconLossHandler)
        # deprecation error
        if normalize_mode is _NO_ARG:
            raise ValueError(
                f'default argument for normalize_mode was "sum", this has been deprecated and will change to "none" in future. Please manually override this value!'
            )
        # load the kernel
        self._kernel = FftKernel(kernel=kernel, normalize_mode=normalize_mode)
        # kernel weighting
        assert 0 <= wrap_weight, f"loss_weight must be in the range [0, inf) but received: {repr(wrap_weight)}"
        assert 0 <= aug_weight, f"kern_weight must be in the range [0, inf) but received: {repr(aug_weight)}"
        self._wrap_weight = wrap_weight
        self._aug_weight = aug_weight
        # disable gradients
        for param in self.parameters():
            param.requires_grad = False

    def activate(self, x_partial: torch.Tensor):
        return self._recon_loss_handler.activate(x_partial)

    def compute_unreduced_loss(self, x_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        wrap_loss = self._recon_loss_handler.compute_unreduced_loss(x_recon, x_targ)
        aug_loss = self._recon_loss_handler.compute_unreduced_loss(self._kernel(x_recon), self._kernel(x_targ))
        return (self._wrap_weight * wrap_loss) + (self._aug_weight * aug_loss)

    def compute_unreduced_loss_from_partial(self, x_partial_recon: torch.Tensor, x_targ: torch.Tensor) -> torch.Tensor:
        return self.compute_unreduced_loss(self.activate(x_partial_recon), x_targ)


# ========================================================================= #
# Registry & Factory                                                        #
# TODO: add ability to register parameterized reconstruction losses
# ========================================================================= #


def _make_aug_recon_loss_l_w_n(loss: str, kern: str, loss_weight: str, kernel_weight: str, normalize_mode: str):
    def _loss(reduction: str):
        return AugmentedReconLossHandler(
            make_reconstruction_loss(loss, reduction=reduction),
            kernel=kern,
            wrap_weight=float(loss_weight),
            aug_weight=float(kernel_weight),
            normalize_mode=normalize_mode,
        )

    return _loss


def _make_aug_recon_loss_l1_w1_n(loss: str, kern: str, normalize_mode: str):
    def _loss(reduction: str):
        return AugmentedReconLossHandler(
            make_reconstruction_loss(loss, reduction=reduction),
            kernel=kern,
            wrap_weight=1.0,
            aug_weight=1.0,
            normalize_mode=normalize_mode,
        )

    return _loss


def _make_aug_recon_loss_l1_w1_nnone(loss: str, kern: str):
    def _loss(reduction: str):
        return AugmentedReconLossHandler(
            make_reconstruction_loss(loss, reduction=reduction),
            kernel=kern,
            wrap_weight=1.0,
            aug_weight=1.0,
            normalize_mode="none",
        )

    return _loss


# NOTE: this function compliments make_kernel in transform/_augment.py
def make_reconstruction_loss(name: str, reduction: str) -> ReconLossHandler:
    return R.RECON_LOSSES[name](reduction=reduction)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from typing import Union

import numpy as np
import torch

# ========================================================================= #
# Generic Helper Functions                                                  #
# - torch, numpy, scalars                                                   #
# - eagerpy is one solution to including this ourselves                     #
# ========================================================================= #


TypeGenericTensor = Union[float, int, torch.Tensor, np.ndarray]
TypeGenericTorch = Union[float, int, torch.Tensor]
TypeGenericNumpy = Union[float, int, np.ndarray]


def generic_as_int32(input: TypeGenericTensor):
    if isinstance(input, (float, int)):
        return int(input)
    elif isinstance(input, torch.Tensor):
        return input.to(torch.int32)
    elif isinstance(input, np.ndarray):
        return input.astype(np.int32)
    else:
        raise TypeError(f"invalid type: {type(input)}")


def generic_max(input: TypeGenericTensor):
    if isinstance(input, (float, int)):
        return input
    elif isinstance(input, torch.Tensor):
        return input.max()
    elif isinstance(input, np.ndarray):
        return input.max()
    else:
        raise TypeError(f"invalid type: {type(input)}")


def generic_min(input: TypeGenericTensor):
    if isinstance(input, (float, int)):
        return input
    elif isinstance(input, torch.Tensor):
        return input.min()
    elif isinstance(input, np.ndarray):
        return input.min()
    else:
        raise TypeError(f"invalid type: {type(input)}")


def generic_shape(input: TypeGenericTensor):
    if isinstance(input, (float, int)):
        return ()
    elif isinstance(input, torch.Tensor):
        return input.shape
    elif isinstance(input, np.ndarray):
        return input.shape
    else:
        raise TypeError(f"invalid type: {type(input)}")


def generic_ndim(input: TypeGenericTensor):
    if isinstance(input, (float, int)):
        return 0
    elif isinstance(input, torch.Tensor):
        return input.ndim
    elif isinstance(input, np.ndarray):
        return input.ndim
    else:
        raise TypeError(f"invalid type: {type(input)}")


# ========================================================================= #
# end                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging

from disent.dataset.data._groundtruth import Hdf5GroundTruthData
from disent.dataset.util.datafile import DataFileHashedDlH5

# ========================================================================= #
# shapes3d                                                                  #
# ========================================================================= #


class Shapes3dData(Hdf5GroundTruthData):
    """
    3D Shapes Dataset:
    - https://github.com/deepmind/3d-shapes

    Files:
        - direct:   https://storage.googleapis.com/3d-shapes/3dshapes.h5
          redirect: https://storage.cloud.google.com/3d-shapes/3dshapes.h5
          info:     https://console.cloud.google.com/storage/browser/_details/3d-shapes/3dshapes.h5
    """

    # TODO: name should be `shapes3d` so that it is a valid python identifier
    name = "3dshapes"

    factor_names = ("floor_hue", "wall_hue", "object_hue", "scale", "shape", "orientation")
    factor_sizes = (10, 10, 10, 8, 4, 15)  # TOTAL: 480000
    img_shape = (64, 64, 3)

    datafile = DataFileHashedDlH5(
        # download file/link
        uri="https://storage.googleapis.com/3d-shapes/3dshapes.h5",
        uri_hash={"fast": "85b20ed7cc8dc1f939f7031698d2d2ab", "full": "099a2078d58cec4daad0702c55d06868"},
        # processed dataset file
        file_hash={"fast": "e3a1a449b95293d4b2c25edbfcb8e804", "full": "b5187ee0d8b519bb33281c5ca549658c"},
        # h5 re-save settings
        hdf5_dataset_name="images",
        hdf5_chunk_size=(1, 64, 64, 3),
        hdf5_obs_shape=img_shape,
    )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #

if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    Shapes3dData(prepare=True)



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import os
from typing import Tuple

import numpy as np
import torch
from torch.utils.data import DataLoader

# ========================================================================= #
# COMPUTE DATASET STATS                                                     #
# ========================================================================= #


@torch.no_grad()
def compute_data_mean_std(
    data,
    batch_size: int = 256,
    num_workers: int = min(os.cpu_count(), 16),
    progress: bool = False,
    chn_is_last: bool = False,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Input data when collected using a DataLoader should return
    `torch.Tensor`s, output mean and std are an `np.ndarray`s
    """
    loader = DataLoader(
        data,
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        drop_last=False,
    )
    if progress:
        from tqdm import tqdm

        loader = tqdm(loader, desc=f"{data.__class__.__name__} stats", total=(len(data) + batch_size - 1) // batch_size)
    # reduction dims
    dims = (1, 2) if chn_is_last else (2, 3)
    # collect obs means & stds
    img_means, img_stds = [], []
    for batch in loader:
        assert isinstance(batch, torch.Tensor), f"batch must be an instance of torch.Tensor, got: {type(batch)}"
        assert batch.ndim == 4, f"batch shape must be: (B, C, H, W), got: {tuple(batch.shape)}"
        batch = batch.to(torch.float64)
        img_means.append(torch.mean(batch, dim=dims))
        img_stds.append(torch.std(batch, dim=dims))
    # aggregate obs means & stds
    mean = torch.mean(torch.cat(img_means, dim=0), dim=0)
    std = torch.mean(torch.cat(img_stds, dim=0), dim=0)
    # checks!
    assert mean.ndim == 1
    assert std.ndim == 1
    # done!
    return mean.numpy(), std.numpy()


# ========================================================================= #
# HELPER                                                                    #
# ========================================================================= #


if __name__ == "__main__":

    def main(progress=True, num_workers=32, batch_size=2048):  # try changing workers to zero on MacOS
        from disent.dataset import data
        from disent.dataset.transform import ToImgTensorF32
        from docs.examples.extend_experiment.code import groundtruth__xyblocks as edat

        for data_cls in [
            # groundtruth -- impl
            data.Cars3dData,
            data.Cars3d64Data,
            data.SmallNorbData,
            data.SmallNorb64Data,
            data.DSpritesData,
            data.Shapes3dData,
            data.SpritesData,
            data.SpritesAllData,
            # groundtruth -- impl synthetic
            data.XYObjectData,
            data.XYObjectShadedData,
            edat.XYBlocksData,
            data.XYSquaresData,
            data.XYSquaresMinimalData,
            data.XColumnsData,
            # groundtruth -- increasing overlap
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=8)),
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=7)),
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=6)),
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=5)),
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=4)),
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=3)),
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=2)),
            (data.XYSquaresData, dict(grid_size=8, grid_spacing=1)),
            (data.XYSquaresData, dict(rgb=False)),
            # large datasets
            (data.Mpi3dData, dict(subset="toy", in_memory=True)),
            (data.Mpi3dData, dict(subset="realistic", in_memory=True)),
            (data.Mpi3dData, dict(subset="real", in_memory=True)),
        ]:
            from disent.dataset.transform import ToImgTensorF32

            # get arguments
            if isinstance(data_cls, tuple):
                data_cls, kwargs = data_cls
            else:
                data_cls, kwargs = data_cls, {}
            # Most common standardized way of computing the mean and std over observations
            # resized to 64px in size of dtype float32 in the range [0, 1].
            data = data_cls(transform=ToImgTensorF32(size=64), **kwargs)
            mean, std = compute_data_mean_std(data, progress=progress, num_workers=num_workers, batch_size=batch_size)
            # results!
            print(
                f"{data.__class__.__name__} - {data.name} - {kwargs}:\n    mean: {mean.tolist()}\n    std: {std.tolist()}"
            )

    # RUN!
    main()


# ========================================================================= #
# RESULTS: 2021-11-12                                                       #
# ========================================================================= #


# Cars3dData - cars3d - {}:
#     mean: [0.8976676149976628, 0.8891658020067508, 0.885147515814868]
#     std: [0.22503195531503034, 0.2399461278981261, 0.24792106319684404]
# DSpritesData - dsprites - {}:
#     mean: [0.042494423521889584]
#     std: [0.19516645880626055]
# SmallNorbData - smallnorb - {}:
#     mean: [0.7520918401088603]
#     std: [0.09563879016827263]
# Shapes3dData - 3dshapes - {}:
#     mean: [0.502584966788819, 0.5787597566089667, 0.6034499731859578]
#     std: [0.2940814043555559, 0.34439790875172144, 0.3661685981524748]

# XYBlocksData - xyblocks - {}:
#     mean: [0.10040509259259259, 0.10040509259259259, 0.10040509259259259]
#     std: [0.21689087652106678, 0.21689087652106676, 0.21689087652106678]
# XYObjectData - xy_object - {}:
#     mean: [0.009818761549013288, 0.009818761549013288, 0.009818761549013288]
#     std: [0.052632363725245844, 0.05263236372524584, 0.05263236372524585]
# XYObjectShadedData - xy_object - {}:
#     mean: [0.009818761549013288, 0.009818761549013288, 0.009818761549013288]
#     std: [0.052632363725245844, 0.05263236372524584, 0.05263236372524585]
# XYSquaresData - xy_squares - {}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920855, 0.12403473458920854, 0.12403473458920854]
# XYSquaresMinimalData - xy_squares_minimal - {}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920855, 0.12403473458920854, 0.12403473458920854]
# XColumnsData - x_columns - {}:
#     mean: [0.125, 0.125, 0.125]
#     std: [0.33075929223788925, 0.3307592922378891, 0.3307592922378892]

# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 8}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920855, 0.12403473458920854, 0.12403473458920854]
# overlap between squares for reconstruction loss, 7 < 8
# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 7}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920854, 0.12403473458920854, 0.12403473458920854]
# overlap between squares for reconstruction loss, 6 < 8
# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 6}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920854, 0.12403473458920854, 0.12403473458920855]
# overlap between squares for reconstruction loss, 5 < 8
# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 5}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920855, 0.12403473458920855, 0.12403473458920854]
# overlap between squares for reconstruction loss, 4 < 8
# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 4}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920855, 0.12403473458920854, 0.12403473458920854]
# overlap between squares for reconstruction loss, 3 < 8
# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 3}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920854, 0.12403473458920854, 0.12403473458920854]
# overlap between squares for reconstruction loss, 2 < 8
# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 2}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920854, 0.12403473458920854, 0.12403473458920854]
# overlap between squares for reconstruction loss, 1 < 8
# XYSquaresData - xy_squares - {'grid_size': 8, 'grid_spacing': 1}:
#     mean: [0.015625, 0.015625, 0.015625]
#     std: [0.12403473458920855, 0.12403473458920855, 0.12403473458920855]
# XYSquaresData - xy_squares - {'rgb': False}:
#     mean: [0.046146392822265625]
#     std: [0.2096506119375896]

# TODO -- REGENERATE:
# Mpi3dData - mpi3d_toy - {'subset': 'toy', 'in_memory': True}:
#     mean: [0.22681593831231503, 0.22353985202496676, 0.22666059934624702]
#     std: [0.07854112062669572, 0.07319301658077378, 0.0790763900050426]
# Mpi3dData - mpi3d_realistic - {'subset': 'realistic', 'in_memory': True}:
#     mean: [0.18240164396358813, 0.20723063241107917, 0.1820551008003256]
#     std: [0.09511163559287175, 0.10128881101801782, 0.09428244469525177]
# Mpi3dData - mpi3d_real - {'subset': 'real', 'in_memory': True}:
#     mean: [0.13111154099374112, 0.16746449372488892, 0.14051725201807627]
#     std: [0.10137409845578041, 0.10087824338375781, 0.10534121043187629]


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
from typing import List
from typing import Tuple

import numpy as np

from disent.dataset.data import BaseEpisodesData
from disent.util.inout.files import download_file
from disent.util.inout.paths import filename_from_url
from disent.util.inout.tar import tar_safe_extract_all

log = logging.getLogger(__name__)


# ========================================================================= #
# custom episodes -- impl                                                   #
# ========================================================================= #


class EpisodesPickledData(BaseEpisodesData):
    def __init__(self, required_file: str, transform=None):
        assert os.path.isabs(required_file), f"{required_file=} must be an absolute path."
        self._required_file = required_file
        # load data
        super().__init__(transform=transform)

    # TODO: convert this to data files?
    # TODO: convert this to data files?
    # TODO: convert this to data files?

    def _load_episode_observations(self) -> List[np.ndarray]:
        import pickle

        # load the raw data!
        with open(self._required_file, "rb") as f:
            # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
            # - Each element in the root list represents an episode
            # - An episode is a list containing many executed options
            # - Each option is a tuple containing:
            #     1. The option name
            #     2. The option id
            #     3. A list of ground truth states covering the option execution. Each ground truth state is a dictionary
            #     4. A list of environment images covering the option execution
            # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
            # Episode = List[Options]
            #   Options = List[Option]
            #     Option  = Tuple[OptionName, OptionId, GroundTruthStates, ObservedStates]
            #       OptionName        = str
            #       OptionId          = int
            #       GroundTruthStates = List[dict]
            #       ObservedStates    = List[np.ndarray]
            # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
            raw_episodes = pickle.load(f)
        # check variables
        option_ids_to_names = {}
        ground_truth_keys = None
        img_shape = None
        # load data
        episodes = []
        for i, raw_episode in enumerate(raw_episodes):
            rollout = []
            for j, raw_option in enumerate(raw_episode):
                # GET: option info
                raw_option: Tuple[str, int, List[dict], List[np.ndarray]]
                option_name, option_id, ground_truth_states, observed_states = raw_option
                # CHECK: number of observations
                assert len(ground_truth_states) == len(observed_states)
                # CHECK: option ids and names
                if option_id not in option_ids_to_names:
                    option_ids_to_names[option_id] = option_name
                else:
                    assert option_ids_to_names[option_id] == option_name
                # CHECK: ground truth keys
                if ground_truth_keys is None:
                    ground_truth_keys = set(ground_truth_states[0].keys())
                else:
                    for gt_state in ground_truth_states:
                        assert ground_truth_keys == gt_state.keys()
                # CHECK: observation shapes
                if img_shape is None:
                    img_shape = observed_states[0].shape
                else:
                    for observation in observed_states:
                        assert observation.shape == img_shape
                # APPEND: all observations into one long episode
                rollout.extend(observed_states)
                # cleanup unused memory! This is not ideal, but works well.
                raw_episode[j] = None
            # make the long episode!
            episodes.append(np.array(rollout))
            # cleanup unused memory! This is not ideal, but works well.
            raw_episodes[i] = None
        # done!
        return episodes


class EpisodesDownloadZippedPickledData(EpisodesPickledData):
    # TODO: convert this to data files?
    # TODO: convert this to data files?
    # TODO: convert this to data files?

    def __init__(self, required_file: str, download_url=None, force_download=False, transform=None):
        self._download_and_extract_if_needed(
            download_url=download_url, required_file=required_file, force_download=force_download
        )
        super().__init__(required_file=required_file, transform=transform)

    def _download_and_extract_if_needed(self, download_url: str, required_file: str, force_download: bool):
        # TODO: this function should probably be moved to the io file.
        # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
        # skip if no download url
        if not isinstance(download_url, str):
            return
        # download file, but skip if file already exists
        save_path = os.path.join(os.path.dirname(required_file), filename_from_url(download_url))
        if force_download or not os.path.exists(save_path):
            log.info(f"Downloading: {download_url=} to {save_path=}")
            download_file(download_url, save_path=save_path)
            log.info(f"Downloaded!")
        # check that the downloaded file exists
        assert os.path.exists(save_path), "The file specified for download does not exist!"
        # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~
        # unzip data
        if (save_path != required_file) and not os.path.exists(required_file):
            if save_path.endswith(".tar.xz"):
                log.info(f"Extracting: {save_path=} to {required_file=}")
                tar_safe_extract_all(
                    save_path,
                    os.path.dirname(required_file),
                )
                log.info(f"Extracted!")
            else:
                raise IOError(f"Unsupported extension for: {save_path}")
        # check that everything exists
        assert os.path.exists(
            required_file
        ), "The required file does not exist after downloading and extracting if necessary!"
        # ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

from dataclasses import dataclass
from numbers import Number
from typing import Any
from typing import Dict
from typing import Sequence
from typing import Tuple
from typing import Union

import torch

from disent.frameworks.ae._unsupervised__ae import Ae
from disent.nn.loss.triplet import TripletLossConfig
from disent.nn.loss.triplet import compute_triplet_loss

# ========================================================================= #
# triple ae                                                                 #
# ========================================================================= #


class TripletAe(Ae):
    REQUIRED_OBS = 3

    @dataclass
    class cfg(Ae.cfg, TripletLossConfig):
        pass

    def hook_ae_compute_ave_aug_loss(
        self, zs: Sequence[torch.Tensor], xs_partial_recon: Sequence[torch.Tensor], xs_targ: Sequence[torch.Tensor]
    ) -> Tuple[Union[torch.Tensor, Number], Dict[str, Any]]:
        return compute_triplet_loss(zs=zs, cfg=self.cfg)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import warnings
from typing import List
from typing import Optional
from typing import Union

import torch

# ========================================================================= #
# Types                                                                     #
# ========================================================================= #


_DimTypeHint = Optional[Union[int, List[int]]]

_POS_INF = float("inf")
_NEG_INF = float("-inf")

_GENERALIZED_MEAN_MAP = {
    "inf": _POS_INF,
    "maximum": _POS_INF,
    "quadratic": 2,
    "arithmetic": 1,
    "geometric": 0,
    "harmonic": -1,
    "minimum": _NEG_INF,
    "-inf": _NEG_INF,
}


# ========================================================================= #
# Generalized Mean Functions                                                #
# ========================================================================= #


def torch_mean_generalized(xs: torch.Tensor, dim: _DimTypeHint = None, p: Union[float, str] = 1, keepdim: bool = False):
    """
    Compute the generalised mean.
    - p is the power

    harmonic mean ≤ geometric mean ≤ arithmetic mean
    - If values have the same units: Use the arithmetic mean.
    - If values have differing units: Use the geometric mean.
    - If values are rates: Use the harmonic mean.
    """
    if isinstance(p, str):
        p = _GENERALIZED_MEAN_MAP[p]
    # compute the specific extreme cases
    if p == _POS_INF:
        return torch.amax(xs, dim=dim, keepdim=keepdim)
    elif p == _NEG_INF:
        return torch.amin(xs, dim=dim, keepdim=keepdim)
    # compute the number of elements being averaged
    if dim is None:
        dim = list(range(xs.ndim))
    n = torch.prod(torch.as_tensor(xs.shape)[dim])
    # warn if the type is wrong
    if p != 1:
        if xs.dtype != torch.float64:
            warnings.warn(
                f"Input tensor to generalised mean might not have the required precision, type is {xs.dtype} not {torch.float64}."
            )
    # compute the specific cases
    if p == 0:
        # geometric mean
        # orig numerically unstable: torch.prod(xs, dim=dim) ** (1 / n)
        return torch.exp((1 / n) * torch.sum(torch.log(xs), dim=dim, keepdim=keepdim))
    elif p == 1:
        # arithmetic mean
        return torch.mean(xs, dim=dim, keepdim=keepdim)
    else:
        # generalised mean
        return ((1 / n) * torch.sum(xs**p, dim=dim, keepdim=keepdim)) ** (1 / p)


def torch_mean_quadratic(xs, dim: _DimTypeHint = None, keepdim: bool = False):
    return torch_mean_generalized(xs, dim=dim, p="quadratic", keepdim=keepdim)


def torch_mean_geometric(xs, dim: _DimTypeHint = None, keepdim: bool = False):
    return torch_mean_generalized(xs, dim=dim, p="geometric", keepdim=keepdim)


def torch_mean_harmonic(xs, dim: _DimTypeHint = None, keepdim: bool = False):
    return torch_mean_generalized(xs, dim=dim, p="harmonic", keepdim=keepdim)


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import os
import os.path

import pytest

import experiment.run as experiment_run
from experiment.util.hydra_main import hydra_main
from tests.util import temp_environ
from tests.util import temp_sys_args

# ========================================================================= #
# TESTS                                                                     #
# ========================================================================= #


EXAMPLE_CFG_DIR = os.path.abspath(os.path.join(__file__, "../..", "docs/examples/extend_experiment/config"))


@pytest.mark.parametrize(
    ("env", "args"),
    [
        # test the standard configs
        (dict(), ["run_action=skip"]),
        (dict(), ["run_action=prepare_data"]),
        (dict(), ["run_action=train"]),
        # test the configs with the research components
        # -- we need to modify the search path
        # -- we need to register all the components
        (
            dict(DISENT_CONFIGS_PREPEND=EXAMPLE_CFG_DIR),
            [
                "run_action=train",
                "dataset=E--xyblocks",
                "metrics=test",
                "framework=E--si-betavae",
                "schedule=adanegtvae_up_all",
            ],
        ),
        (
            dict(DISENT_CONFIGS_PREPEND=EXAMPLE_CFG_DIR),
            [
                "run_action=train",
                "dataset=E--mask-dthr-pseudorandom.yaml",
                "metrics=none",
                "framework=adavae_os",
                "schedule=beta_cyclic",
            ],
        ),
    ],
)
def test_experiment_run(env, args):
    # show full errors in hydra
    os.environ["HYDRA_FULL_ERROR"] = "1"

    # temporarily set the environment and the arguments
    with temp_environ(env), temp_sys_args([experiment_run.__file__, *args]):
        # run the hydra experiment
        # 1. sets the default search path to `experiment/config`
        # 2. add to the search path with the `DISENT_CONFIGS_PREPEND` and `DISENT_CONFIGS_APPEND` environment variables
        # 3. enable the ${exit:<msg>} and various other resolvers for omegaconf/hydra
        hydra_main(
            callback=experiment_run.run_action,
            config_name="config_test",
            log_level=logging.DEBUG,
        )


# ========================================================================= #
# END                                                                       #
# ========================================================================= #



#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
#  MIT License
#
#  Copyright (c) 2021 Nathan Juraj Michlo
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy
#  of this software and associated documentation files (the "Software"), to deal
#  in the Software without restriction, including without limitation the rights
#  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#  copies of the Software, and to permit persons to whom the Software is
#  furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
#  SOFTWARE.
#  ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~

import logging
import time
from typing import Optional

import lightning as L

log = logging.getLogger(__name__)


# ========================================================================= #
# HELPER CALLBACKS                                                          #
# ========================================================================= #


class BaseCallbackPeriodic(L.Callback):
    def __init__(self, every_n_steps: Optional[int] = None, begin_first_step: bool = False):
        assert (every_n_steps is None) or (
            isinstance(every_n_steps, int) and every_n_steps > 0
        ), f"`every_n_steps` must be None or an integer greater than zero, got: {repr(every_n_steps)}"
        self.every_n_steps = every_n_steps
        self.begin_first_step = begin_first_step

    def on_train_start(self, trainer: L.Trainer, pl_module: L.LightningModule):
        if self.every_n_steps is None:
            # number of steps/batches in an epoch
            self.every_n_steps = trainer.num_training_batches

    def on_batch_end(self, trainer: L.Trainer, pl_module: L.LightningModule):
        if 0 == trainer.global_step % self.every_n_steps:
            # skip on the first step if required
            if trainer.global_step == 0 and not self.begin_first_step:
                return
            self.do_step(trainer, pl_module)

    def do_step(self, trainer: L.Trainer, pl_module: L.LightningModule):
        raise NotImplementedError


class BaseCallbackTimed(L.Callback):
    def __init__(self, interval: float = 10):
        assert interval > 0, f"The interval must be greater than zero, got: {repr(interval)}"
        self._last_time = 0
        self._interval = interval
        self._start_time = time.time()

    def on_batch_end(self, trainer: L.Trainer, pl_module: L.LightningModule):
        current_time = time.time()
        step_delta = current_time - self._last_time
        # only run every few seconds
        if self._interval < step_delta:
            self._last_time = current_time
            self.do_interval(trainer, pl_module, current_time, self._start_time)

    def do_interval(self, trainer: L.Trainer, pl_module: L.LightningModule, current_time, start_time):
        raise NotImplementedError


# ========================================================================= #
# END                                                                       #
# ========================================================================= #
